
/****** move.c **********************************************************/
/*									*/
/*   Included in this module are two recursive coordinate bisection	*/
/*   program based on the median-of-medians algorithm. The programs	*/
/*   move the data based on the partitioning.				*/
/*                                                                      */
/*   RecCoorBisec:       Recursive coordinate bisection program.	*/
/*   WeightRecCoorBisec: Weighted recursive coordinate bisection.	*/
/*   									*/
/*************************************************** Shin 11/22/93 ******/

#include <stdio.h>
#include "parti.h"
#include "bisec.h"

#define  MAXPART	512	/* maximum number of partitions */

extern	double	*doublelist;	/* double array */

/* recursive coordinate bisection */
void
RecCoorBisec (level, remaplevel, myindex, ndata, ndim, data)
int    level;			/* level of binary disection */
int    remaplevel;		/* remap for every remaplevel */
int    *myindex;		/* current indexes local data */
int    *ndata;			/* number of nodes */
int    ndim;			/* number of dimensions */
double *data[];    		/* coordinate arrays */

{
    int    curndata;		/* cursnt size */
    int    totalpart;		/* total number of partitions */
    int    npart;		/* current number of partitions */
    int    direction;		/* partitioning direction */
    int    parlevel;		/* level of each partition */
    int    parproc;		/* # of processors of each partition */
    int    mask;		/* mask */
    int    ptr[MAXPART+1];	/* pointer of begin of partition */
    int    offset[MAXPART];	/* offset within partitions */
    int    *proc;		/* processor assignment */
    int    *index;		/* index of arrays */
    int    i, j, k, n, no;
    Val    *midlist;		/* array of medians */
    Val    *temp;
    SCHED  *sched;		/* schedule for remapping */
    TTABLE *table;		/* translation table */
    TTABLE *init_ttable_with_proc();

    /* check the remapping level */
    if (remaplevel <= 0) {
	printf ("Remapping level must be greater than 0.\n");
	exit(1);
    }

    /* initialize parameters */
    curndata = *ndata;
    totalpart = (1 << remaplevel);

    /* allocate memory for necessary arrays*/
    index   = (int *) malloc ((curndata<<1) * sizeof(int));
    proc    = (int *) malloc ((curndata<<1) * sizeof(int));
    midlist = (Val *) malloc (totalpart * PARTI_numnodes * sizeof(Val));
    temp    = (Val *) malloc (totalpart * PARTI_numnodes * sizeof(Val));

    /* put the coordinates into data array */
    for (i = 0; i <= curndata; i++) {
        index[i] = i;
	proc[i] = 0;
    }

    /* sort the arrays of each partition and find the median */
    npart = 1;				/* number of partitions */
    for (i=0, ptr[0]=0, parlevel=level; i < level; i++) {
	/* remap all data if level is equal to remaplevel */
	if ((i / remaplevel) && ((i % remaplevel) == FALSE)) {
	    /* calculate the number of processors of each partition */
	    parlevel -= remaplevel;		/* level for global func */
	    parproc = 1 << parlevel;		/* # of processors */
	    mask = parproc - 1;			/* mask for proc */

	    /* set an offset array for distribution */
	    for (j = 0; j < npart; j++)
		offset[j] = PARTI_mynode & mask;

	    /* set up the proc array */
	    for (j = 0; j < curndata; j++) {
		offset[proc[j]] &= mask;
		proc[j] = (proc[j] << parlevel) + offset[proc[j]]++;
	    }

	    /* output the mapping array
	    for (j = 0; j < PARTI_numnodes; j++) {
	        if (PARTI_mynode == j) {
		    pfd = fopen ("map", "a+");
		    for (k = 0; k < curndata; k++)
		        fprintf(pfd, "%d\n", proc[k]);
		    fclose (pfd);
	        }
	        gsync();
	    }	*/

	    /* build translation table */
	    table = init_ttable_with_proc (1, proc, curndata);

	    /* call remap function to build a schedule */
	    remap (table, myindex, &sched, myindex, &curndata);

	    /* move the x, y, and z arrays */
#ifdef	MULTI_GATHER
	    PARTI_multi_gather (sched, sizeof(double), 3, data, data);
#else
	    for (j = 0; j < ndim; j++)
	    	PARTI_gather (sched, data[j], data[j], sizeof(double));
#endif

	    /* free the schedule */
	    free_sched (sched);
	    free_table (table);

	    /* reset the proc array */
	    for (j = 0; j <= curndata; j++) {
		proc[j] = PARTI_mynode >> parlevel;
		index[j] = j;
	    }

	    /* number of partition per subcube */
	    npart = 1;
	}

	/* set direction and size of partitions */
        direction = i % ndim;		/* direction of partition */
        ptr[npart] = curndata;    /* number of nodes */

        /* sort the arrays of each partition */
	doublelist = data[direction];
        for (j = 0; j < npart; j++) {
            /* sort the array via coordinates */
            qsort (&index[ptr[j]], ptr[j+1]-ptr[j], sizeof(int), comparedouble);
        }

	/* apply the method median of medians to find
	   the median of the distributed array */
	find_median (parlevel, npart, ptr, doublelist, index, proc, midlist, temp);

	/* number of partitions */
	npart <<= 1;
    }

    /* build translation table */
    table = init_ttable_with_proc (1, proc, curndata);

    /* call remap function to build a schedule */
    remap (table, myindex, &sched, myindex, ndata);

    /* move the x, y, and z arrays */
#ifdef	MULTI_GATHER
    PARTI_multi_gather (sched, sizeof(double), 3, data, data);
#else
    for (j = 0; j < ndim; j++)
	PARTI_gather (sched, data[j], data[j], sizeof(double));
#endif

    /* free the schedule */
    free_sched (sched);

    /* release memory */
    free (index);
    free (proc);
    free (temp);
    free (midlist);
}

/* weighted recursive coordinate bisection */
WeighRecCoorBisec (level, remaplevel, myindex, load, ndata, ndim, data)
int    level;			/* level of binary disection */
int    remaplevel;		/* remap for every remaplevel */
int    *myindex;		/* current indexes local data */
int    *load;			/* load of every vertex */
int    *ndata;			/* number of nodes */
int    ndim;			/* number of dimensions */
double *data[];    		/* coordinate arrays */

{
    int    curndata;		/* cursnt size */
    int    totalpart;		/* total number of partitions */
    int    npart;		/* current number of partitions */
    int    direction;		/* partitioning direction */
    int    parlevel;		/* level of each partition */
    int    parproc;		/* # of processors of each partition */
    int    mask;		/* mask */
    int    ptr[MAXPART+1];	/* pointer of begin of partition */
    int    offset[MAXPART];	/* offset within partitions */
    int    *proc;		/* processor assignment */
    int    *index;		/* index of arrays */
    int    *weight;		/* weight */
    int    i, j, k, n, no;
    Val    *midlist;		/* array of medians */
    Val    *temp;
    SCHED  *sched;		/* schedule for remapping */
    TTABLE *table;		/* translation table */
    TTABLE *init_ttable_with_proc();
    FILE   *output;

    /* check the remapping level */
    if (remaplevel <= 0) {
	printf ("Remapping level must be greater than 0.\n");
	exit(1);
    }

    /* initialize parameters */
    totalpart = (1 << remaplevel);
    curndata = *ndata;

    /* allocate memory for necessary arrays*/
    index   = (int *) malloc ((curndata<<1) * sizeof(int));
    proc    = (int *) malloc ((curndata<<1) * sizeof(int));
    midlist = (Val *) malloc (totalpart * PARTI_numnodes * sizeof(Val));
    temp    = (Val *) malloc (totalpart * PARTI_numnodes * sizeof(Val));
    /* statements for weighted partitioner */
    weight  = (int *) malloc ((curndata<<1) * sizeof(int));

    /* put the coordinates into data array */
    for (i = 0; i <= curndata; i++) {
	index[i] = i;
	proc[i] = 0;
    }

    /* sort the arrays of each partition and find the median */
    npart = 1;
    for (i=0, ptr[0]=0, parlevel=level; i < level; i++) {
	/* remap all data if level is equal to remaplevel */
        if ((i / remaplevel) && ((i % remaplevel) == FALSE)) {
            /* calculate the number of processors of each partition */
            parlevel -= remaplevel;             /* level for global func */
            parproc = 1 << parlevel;            /* # of processors */
            mask = parproc - 1;                 /* mask for proc */

            /* set an offset array for distribution */
            for (j = 0; j < npart; j++)
                offset[j] = PARTI_mynode & mask;

            /* set up the proc array */
            for (j = 0; j < curndata; j++) {
                offset[proc[j]] &= mask;
                proc[j] = (proc[j] << parlevel) + offset[proc[j]]++;
            }

            /* output the mapping array
            for (j = 0; j < PARTI_numnodes; j++) {
                if (PARTI_mynode == j) {
                    pfd = fopen ("map", "a+");
                    for (k = 0; k < curndata; k++)
                        fprintf(pfd, "%d\n", proc[k]);
                    fclose (pfd);
                }
                gsync();
            } */

            /* build translation table */
            table = init_ttable_with_proc (1, proc, curndata);

            /* call remap function to build a schedule */
            remap (table, myindex, &sched, myindex, &curndata);

            /* move the x, y, z, and load arrays */
#ifdef	MULTI_GATHER
	    PARTI_multi_gather (sched, sizeof(double), 3, data, data);
#else
	    for (j = 0; j < ndim; j++)
            	PARTI_gather (sched, data[j], data[j], sizeof(double));
#endif
            PARTI_gather (sched, load, load, sizeof(int));

            /* free the schedule */
            free_sched (sched);
	    free_table (table);

            /* reset the proc array */
            for (j = 0; j <= curndata; j++) {
                proc[j] = PARTI_mynode >> parlevel;
                index[j] = j;
            }

            /* number of partition per subcube */
            npart = 1;
        }

        /* set direction and size of partitions */
        direction = i % ndim;    /* direction of partition */
        ptr[npart] = curndata;    /* number of nodes */

        /* sort the arrays of each partition */
	doublelist = data[direction];
        for (j = 0; j < npart; j++) {
            /* sort the array via coordinates */
            qsort (&index[ptr[j]], ptr[j+1]-ptr[j], sizeof(int), comparedouble);
        }

	/* install the weight array */
	for (j = 1, weight[0] = 0; j <= curndata; j++)
	    weight[j] = weight[j-1] + load[index[j-1]];

	/* apply the method median of medians to find
	   the median of the distributed array */
	find_median_weight (parlevel, npart, ptr, doublelist, weight, index, proc, midlist, temp);

	/* number of partitions */
        npart <<= 1;
    }

    /* build translation table */
    table = init_ttable_with_proc (1, proc, curndata);

    /* call remap function to build a schedule */
    remap (table, myindex, &sched, myindex, ndata);

    /* move the x, y, and z arrays */
#ifdef	MULTI_GATHER
    PARTI_multi_gather (sched, sizeof(double), 3, data, data);
#else
    for (j = 0; j < ndim; j++)
    	PARTI_gather (sched, data[j], data[j], sizeof(double));
#endif
    PARTI_gather (sched, load, load, sizeof(int));

    /* free the schedule */
    free_sched (sched);
    free_table (table);

    /* restore the load array
    for (i = 1; i < *ndata; i++)
	load[i] += load[i-1]; */

    /* release memory */
    free (index);
    free (proc);
    free (temp);
    free (midlist);
    free (weight);
}

